import sqlite3, pathlib, time, collections

class DB (object) :
    def __init__ (self, path) :
        self.p = pathlib.Path(path).with_suffix("")
        self.p.mkdir(parents=True, exist_ok=True)
        self.d = sqlite3.connect(self.p.with_suffix(".db"))
        self.c = self.d.cursor()
        self.runs_cols = ("date", "project", "test",
                          "prep_ret", "prep_out", "prep_err",
                          "build_ret", "build_out", "build_err",
                          "run_ret", "run_out", "run_err", "run_sys", "run_ass")
        self._runs_row = collections.namedtuple("row", self.runs_cols)
        self._runs = ",".join("?" for c in self.runs_cols)
        self.checks_cols = ("date", "project", "test", "result")
        self._checks_row = collections.namedtuple("row", self.checks_cols)
        self._checks = ",".join("?" for c in self.checks_cols)
        self._create()
    def __del__ (self) :
        try :
            self.d.close()
        except :
            pass
    def close (self) :
        self.d.close()
    def _create (self) :
        todo = {"runs", "checks"}
        rows = self.c.execute("select name from sqlite_master WHERE type='table'")
        todo.difference_update(r[0] for r in rows.fetchall())
        for table in todo :
            cols = ",".join(getattr(self, f"{table}_cols"))
            self.c.execute(f"CREATE TABLE {table} ({cols})")
        if todo :
            self.d.commit()
    def _add (self, table, *cols) :
        _cols = getattr(self, f"{table}_cols")
        assert len(_cols) == len(cols)
        _patt = getattr(self, f"_{table}")
        try :
            self.c.execute(f"INSERT INTO {table} VALUES ({_patt})", cols)
            self.d.commit()
        except :
            self.d.rollback()
            raise
    def _rows (self, table) :
        rtype = getattr(self, f"_{table}_row")
        rows = self.c.execute(f"SELECT * FROM {table} ORDER by date")
        while True :
            r = rows.fetchone()
            if r is None :
                break
            yield rtype(*r)
    def runs (self) :
        return self._rows("runs")
    def checks (self) :
        return self._rows("checks")
    def add_run (self, project, test, run) :
        now = time.strftime("%Y-%m-%d %H:%M:%S")
        now_d, now_t = now.split()
        dup = 0
        base_dir = pathlib.Path(self.p
                                / now_d
                                / "-".join(now_t.split(":")[:2])
                                / f"{project}.{dup}")
        while base_dir.exists() :
            dup += 1
            base_dir = base_dir.parent / f"{project}.{dup}"
        base_dir.mkdir(parents=True)
        cols = [now, project, test]
        for name in ["prep", "build", "run"] :
            for ext in ["ret", "out", "err", "sys", "ass"] :
                if name != "run" and ext in ("sys", "ass") :
                    continue
                key = f"{name}.{ext}"
                val = run.get(key, None)
                if val is None :
                    cols.append(None)
                elif isinstance(val, (int, float)) :
                    cols.append(str(val))
                elif isinstance(val, (list, tuple, dict)) :
                    cols.append(str(base_dir / key))
                    with open(cols[-1], "w") as f :
                        f.write(repr(val))
                elif isinstance(val, str) :
                    cols.append(str(base_dir / key))
                    with open(cols[-1], "w") as f :
                        f.write(val)
                elif callable(getattr(val, "write", None)) :
                    cols.append(str(base_dir / key))
                    with open(cols[-1], "w") as f :
                        val.write(f)
                else :
                    raise ValueError(f"cound not save {key} = {val!r}")
        self._add("runs", *cols)
    def add_check (self, project, test, result) :
        self._add("checks", time.strftime("%Y-%m-%d %H:%M:%S"),
                  project, test, str(result))



import signal
from pathlib import Path
from subprocess import Popen, STDOUT, PIPE
from tempfile import NamedTemporaryFile
from collections import namedtuple

class SEvent (object) :
    def __init__ (self, pid, time, kind, **info) :
        self.pid = pid
        self.time = time
        self.kind = kind
        self._info = info
        for key, val in info.items() :
            setattr(self, key, val)
    def __repr__ (self) :
        return (f"{self.__class__.__name__}({self.pid}, {self.time:f}, {self.kind!r}, "
                + ", ".join(f"{k}={v!r}" for k, v in self._info.items())
                + ")")

call = namedtuple("call", ["pid", "time", "head", "resume"])

class STrace (object) :
    def __init__ (self) :
        self.pids = {}
        self.root = None
    def write (self, stream) :
        stream.write(repr((self.root, self.pids)))
    def _print (self, out) :
        for pid, events in sorted(self.pids.items()) :
            if pid == self.root :
                out.write(f"> {pid} (parent)\n")
            else :
                out.write(f"> {pid} (child)\n")
            for evt in self.pids[pid] :
                out.write(f"  {evt}\n")
    @classmethod
    def read (cls, stream) :
        self = cls()
        self.root, self.pids = eval(stream.read())
    @classmethod
    def parse (cls, lines) :
        self = cls()
        started = {}
        for pid, time, info in (l.strip().split(None, 2) for l in lines) :
            if info.endswith("<unfinished ...>") :
                name = info.split("(", 1)[0]
                started[pid] = call(pid, time, info[:-16], f"<... {name} resumed>")
                continue
            elif pid in started :
                assert info.startswith(started[pid].resume), "wrong resume"
                start = started.pop(pid)
                time, info = start.time, start.head + info[len(start.resume):]
            pid, time = int(pid), float(time)
            if self.root is None :
                self.root = pid
            if pid not in self.pids :
                self.pids[pid] = []
            if info.startswith("+++") :
                status = info.split()[3]
                try :
                    code = int(status)
                    evt = "exit"
                except :
                    if hasattr(signal, status) :
                        code = status
                        evt = "killed"
                    else :
                        code = status
                        evt = "died"
                self.pids[pid].append(SEvent(pid, time, evt, code=code))
            elif info.startswith("---") :
                info = info.strip("- ")
                sig, info = info.split(None, 1)
                params = dict(x.split("=", 1) for x in info.strip("{}").split(", "))
                self.pids[pid].append(SEvent(pid, time, "signal",
                                             name=sig,
                                             params=params))
            else :
                name = info.split("(", 1)[0]
                params, ret = info[len(name):].rsplit("=", 1)
                self.pids[pid].append(SEvent(pid, time, "call",
                                             name=name.strip(),
                                             params=params.strip(),
                                             ret=ret.strip()))
        return self

class AssRunner (object) :
    def __init__ (self, add, rep) :
        self.add = [Path(p).absolute() for p in add]
        self.rep = [Path(p).absolute() for p in rep]
    def __call__ (self, path, timeout=60) :
        proj = Path(path)
        script = []
        script.extend(f"cp -f '{p}' ." for p in self.rep)
        script.extend(f"cp -f '{p}' ." for p in self.add
                      if not (proj / p.name).exists())
        script.extend(["touch run.ass run.sys argv",
                       "sh prepare.sh >prep.out 2>prep.err",
                       "echo $? > prep.ret",
                       "sh build.sh >build.out 2>build.err",
                       "echo $? > build.ret",
                       "strace -f -r -o run.sys ./a.out $(cat argv) >run.out 2>run.err </dev/null",
                       "echo $? > run.ret",
                       "echo '$' >> prep.out",
                       "echo '$' >> prep.err",
                       "echo '$' >> build.out",
                       "echo '$' >> build.err",
                       "echo '$' >> run.out",
                       "echo '$' >> run.err",
                       "set -x",
                       "cat prep.ret",
                       "cat prep.out",
                       "cat prep.err",
                       "cat build.ret",
                       "cat build.out",
                       "cat build.err",
                       "cat run.ret",
                       "cat run.out",
                       "cat run.err",
                       "cat run.ass",
                       "cat run.sys"])
        with NamedTemporaryFile(mode="w", dir=proj, suffix=".sh") as tmp :
            tmp.write("\n".join(script) + "\n")
            tmp.flush()
            proc = Popen(["firejail",
                          "--quiet",
                          "--overlay-tmpfs",
                          "--allow-debuggers",
                          f"--rlimit-cpu={timeout}",
                          "sh", tmp.name],
                         cwd=proj,
                         stdout=PIPE,
                         stderr=STDOUT,
                         encoding="utf-8")
            proc.wait()
        run = {}
        buf = []
        for line in (l.rstrip() for l in proc.stdout) :
            if line.startswith("+ cat ") :
                buf = run[line.split()[-1]] = []
            else :
                buf.append(line)
        for key, val in run.items() :
            if key == "run.ass" :
                log = {}
                for line in val :
                    ret, tag, test = line[2:-2].split(":", 2)
                    log[tag] = (ret, test)
                run[key] = log
            elif key.endswith(".out") or key.endswith(".err") :
                assert(val[-1].endswith("$"))
                val[-1] = val[-1][:-1]
                run[key] = "\n".join(val).rstrip()
            elif key.endswith(".ret") :
                run[key] = int(val[0].strip())
            elif key.endswith(".sys") :
                run[key] = STrace.parse(val)
        return run



import argparse, pathlib, importlib, sys, functools, ast
import lxml.etree
import pandas as pd
from colorama import init as colorama_init, Fore, Style
from .db import DB
from .p5 import PrePreProcessor

colorama_init()

def record (method) :
    @functools.wraps(method)
    def wrapper (self, args) :
        ret = method(self, args)
        if args.record :
            self.log.write(f"{Fore.BLUE}{Style.BRIGHT}save{Style.RESET_ALL}"
                           f" {args.record} in {args.db}\n")
            project, test = args.record.split(":", 1)
            db = DB(args.db)
            if method.__name__ == "do_run" :
                db.add_run(project, test, ret)
            else :
                db.add_check(project, test, ret)
            db.close()
        return ret
    return wrapper

def _const (val) :
    if isinstance(val, str) :
        val = ast.literal_eval(val)
    if isinstance(val, bool) :
        return int(val)
    else :
        return val

class BadassCLI (object) :
    def __init__ (self, path, out, log, lang) :
        self.path = pathlib.Path(path)
        self.out = out
        self.log = log
        self.lang = lang.lower()
        self.mod = importlib.import_module("." + self.lang, "badass")
        self.src = {}
    def __call__ (self, args) :
        cmd = getattr(self, "do_" + args.command)
        cmd(args)
    def _source_load (self, path, source=None) :
        self.log.write(f"{Fore.BLUE}{Style.BRIGHT}load{Style.RESET_ALL} {path}\n")
        path = (self.path / pathlib.Path(".//" + str(path))).relative_to(self.path)
        if path in self.src :
            return self.src[path]
        src = self.src[path] = self.mod.source.Source(str(self.path / path), source)
        return src
    ##
    ## patch
    ##
    @record
    def do_patch (self, args) :
        base = self._source_load(args.path)
        for cmd, arg in zip(args.expr[::2], args.expr[1::2]) :
            getattr(self, "_do_patch_" + cmd)(base, arg)
        src = base.src()
        self.out.write(src)
        return src
    def _do_patch_del (self, base, name) :
        self.log.write(f" {Fore.RED}{Style.BRIGHT}---{Style.RESET_ALL} {name}\n")
        del base[name]
    def _do_patch_add (self, base, name_path) :
        name, path = name_path.split("@", 1)
        self.log.write(f" {Fore.GREEN}{Style.BRIGHT}+++{Style.RESET_ALL}"
                       f" {name}{Fore.WHITE}{Style.DIM}@{Style.RESET_ALL}{path}\n")
        code = self._source_load(path)
        base.add(code[name])
    ##
    ## xpath
    ##
    @record
    def do_xpath (self, args) :
        if not (args.source or args.xml or args.count) :
            args.source = True
        src = self._source_load(args.path)
        if args.source or args.xml :
            if args.source :
                ret = src(args.expr, "src")
            elif args.xml :
                ret = src(args.expr, "xml")
            if not isinstance(ret, list) :
                ret = [ret]
            if args.xml :
                for i, r in enumerate(ret) :
                    try :
                        ret[i] = lxml.etree.tostring(r, pretty_print=True,
                                                     encoding="utf-8")
                    except :
                        ret[i] = str(r)
            for i, r in enumerate(ret) :
                if i :
                    self.out.write("\n\n")
                if isinstance(r, bytes) :
                    ret[i] = r.decode("utf-8")
                elif not isinstance(r, str) :
                    ret[i] = str(r)
            val = "\n\n".join(r + "\n" for r in ret)
            self.out.write(val)
            return val
        elif args.count :
            ret = src(args.expr)
            if isinstance(ret, list) :
                self.out.write(f"{len(ret)}\n")
                return len(ret)
            elif ret is None :
                self.out.write("0\n")
                return 0
            else :
                self.out.write("1\n")
                return 1
    ##
    ## has
    ##
    @record
    def do_has (self, args) :
        patt, path = args.decl.split("@", 1)
        self._csrc = getattr(self, "_csrc", 0) + 1
        decl = self._source_load(f".{self._csrc}.c", patt + ";")
        name = next(iter(decl.d))
        try :
            code = self._source_load(path)
            if decl.sig(name) == name :
                found = name in code.d
            else :
                found = name in code.d and decl.sig(name) == code.sig(name)
        except FileNotFoundError :
            found = False
        if found :
            patt = code.sig(name)
            path, line, _ = code.loc(name)
            self.log.write(f"{Fore.GREEN}{Style.BRIGHT}found{Style.RESET_ALL}"
                           f" {patt}{Fore.WHITE}{Style.DIM}@{Style.RESET_ALL}"
                           f"{path}:{line}\n")
        else :
            self.log.write(f"{Fore.RED}{Style.BRIGHT}not found{Style.RESET_ALL}"
                           f" {patt}{Fore.WHITE}{Style.DIM}@{Style.RESET_ALL}"
                           f"{path}\n")
        self.out.write(f"{int(found)}\n")
        return found
    ##
    ## cparse
    ##
    def do_cparse (self, args) :
        self._csrc = getattr(self, "_csrc", 0) + 1
        src = self._source_load(f".{self._csrc}.c", args.code)
        xml = lxml.etree.tostring(src.x, pretty_print=True, encoding="utf-8")
        self.out.write(xml.decode("utf-8"))
        self.out.write("\n")
    ##
    ## run
    ##
    _runcolor = {"out" : Fore.GREEN,
                 "err" : Fore.RED,
                 "sys" : Fore.MAGENTA,
                 "ret" : Fore.BLUE,
                 "*"   : Fore.YELLOW}
    @record
    def do_run (self, args) :
        self.log.write(f"{Fore.BLUE}{Style.BRIGHT}run{Style.RESET_ALL} {args.base}\n")
        add, rep = set(), set()
        for a in args.add :
            add.update(a)
        for a in args.replace :
            rep.update(a)
        runner = self.mod.run.AssRunner(add, rep)
        run = runner(args.base, timeout=args.timeout)
        for key, val in run.items() :
            if not getattr(args, key.rsplit(".", 1)[-1], None) :
                continue
            elif val is None or (isinstance(val, str) and not val) :
                continue
            color = self._runcolor.get(key.split(".", 1)[-1], self._runcolor["*"])
            self.out.write(f"{color}=== {key} ==={Style.RESET_ALL}\n")
            if isinstance(val, str) :
                self.out.write(val)
            elif getattr(val, "_print", None) :
                getattr(val, "_print")(self.out)
            else :
                self.out.write(repr(val))
            self.out.write("\n")
        return run
    ##
    ## report
    ##
    def do_report (self, args) :
        self.log.write(f"{Fore.BLUE}{Style.BRIGHT}save{Style.RESET_ALL} {args.outfile}\n")
        db = DB(args.db)
        data = {}
        for row in db.checks() :
            data.setdefault(row.project, {})
            data[row.project][row.test] = _const(row.result)
        for row in db.runs() :
            data.setdefault(row.project, {})
            ass = ast.literal_eval(open(row.run_ass).read())
            for key, (result, _) in ass.items() :
                data[row.project][f"{row.test}_{key}"] = _const(result == "passed")
            data[row.project][f"{row.test}_build"] = _const(row.build_ret == "0")
        df = pd.DataFrame.from_dict(data, orient="index")
        df.to_csv(args.outfile, index=True, index_label="project")
    ##
    ## p5
    ##
    def do_p5 (self, args) :
        g = dict(a.split("=", 1) for a in args.let)
        p = PrePreProcessor(*args.input,
                            comment=args.marker,
                            include=args.include,
                            **g)
        if args.save :
            p.save(args.save)
        if args.nocpp :
            args.output.write(p.ppp)
        else :
            args.output.write(p.cpp)

##
## CLI
##

parser = argparse.ArgumentParser("badass",
                                 description="(not so) bad assessments")
parser.add_argument("-o", "--output", default=sys.stdout,
                    type=argparse.FileType("w", encoding="utf-8"),
                    metavar="PATH",
                    help="print results to PATH (default: stdout)")
parser.add_argument("-l", "--log", default=sys.stderr,
                    type=argparse.FileType("w", encoding="utf-8"),
                    metavar="PATH",
                    help="print log messages to PATH (default: stderr)")
parser.add_argument("--lang", type=str, default="C",
                    help="programming language for the project (default: 'C')")
parser.add_argument("--db", default="bad.db", metavar="PATH",
                    help="PATH to database for result records (default 'bad.db')")
parser.add_argument("--record", default=None, type=str, metavar="PROJECT:TEST",
                    help="record result as TEST in PROJECT (two arbitrary tags)")
parser.add_argument("-b", "--base", metavar="DIR", default=".",
                    help="path of project base directory (default: '.')")
sub = parser.add_subparsers(dest="command",
                            required=True,
                            title="available commands")

patch = sub.add_parser("patch",
                       help="patch a file")
patch.add_argument("path",
                   help="file to be patched")
patch.add_argument("expr",
                   nargs="+",
                   help="patch directives (del NAME / add PATH@NAME)")

xpath = sub.add_parser("xpath",
                       help="query a file using xpath")
fmt = xpath.add_mutually_exclusive_group()
fmt.add_argument("-s", "--source", action="store_true", default=False,
                 help="output result as source code (default)")
fmt.add_argument("-x", "--xml", action="store_true", default=False,
                 help="output result as XML")
fmt.add_argument("-c", "--count", action="store_true", default=False,
                 help="output result as the number of matches")
xpath.add_argument("path",
                   help="file to be searched in")
xpath.add_argument("expr",
                   help="xpath expression")

has = sub.add_parser("has",
                     help="check for a declaration")
has.add_argument("decl",
                 help="declaration to be searched for (name or signature)")

cparse = sub.add_parser("cparse",
                        help="parse C code and dump it as XML")
cparse.add_argument("code", help="C code")

run = sub.add_parser("run",
                     help="build and execute a project")
run.add_argument("-t", "--timeout", default=60, type=int,
                 help="limit of CPU runtime, in seconds (default 60)")
run.add_argument("--out", default=False, action="store_true",
                 help="show stdout")
run.add_argument("--err", default=False, action="store_true",
                 help="show stderr")
run.add_argument("--ret", default=False, action="store_true",
                 help="show return codes")
run.add_argument("--sys", default=False, action="store_true",
                 help="show strace")
run.add_argument("-a", "--add", metavar="FILE",
                 nargs="+", action="append", default=[],
                 help="files to add to the project (unless already present)")
run.add_argument("-r", "--replace", metavar="FILE",
                 nargs="+", action="append", default=[],
                 help="files to replace in the project (or add if absent)")

report = sub.add_parser("report",
                        help="dump database to CSV")
report.add_argument("outfile", default="badass.csv", type=str, nargs="?",
                    help="file name to be saved (default 'badass.csv')")

ppppp = sub.add_parser("p5",
                       help="Python-powered pre-pre-processor")
ppppp.add_argument("-I", metavar="DIR", action="append", default=[], dest="include",
                   help="add DIR the the set of paths searched for includes")
ppppp.add_argument("-m", "--marker", default="//", type=str,
                   help="directives marker (default: '//')")
ppppp.add_argument("-l", "--let", metavar="NAME=EXPR", default=[], action="append",
                   type=str, help="define additional names, as with let directive")
ppppp.add_argument("-s", "--save", metavar="FILE", default=None,
                   type=argparse.FileType("w", encoding="utf-8"),
                   help="save ppppped text to FILE")
ppppp.add_argument("-n", "--nocpp", default=False, action="store_true",
                   help="output text whithout paassing it to cpp")
ppppp.add_argument("input", metavar="FILE", nargs="+",
                   type=argparse.FileType("r", encoding="utf-8"),
                   help="input FILEs to process")

def main () :
    args = parser.parse_args()
    badcli = BadassCLI(args.base, args.output, args.log, args.lang)
    badcli(args)



import subprocess
from lxml import etree
import chardet

from clang.cindex import Index, Config, CursorKind
if not getattr(Config, "library_file", None) :
    Config.set_library_file("/usr/lib/llvm-9/lib/libclang.so")

_cpp_head = ""

def cpp (src, env={}, **var) :
    _env = dict(env)
    _env.update(var)
    src = subprocess.run(["cpp", "-P", "--traditional", "-C"]
                         + [f"-D{k}={v}" for k, v in _env.items()],
                         input=src, encoding="utf-8", capture_output=True).stdout
    return src.replace(_cpp_head, "")

_cpp_head = cpp("")

class Source (object) :
    def __init__ (self, path, source=None) :
        self.p = path
        self.i = Index.create()
        if source is None :
            data = open(path, "rb").read()
            enc = chardet.detect(data)
            self.s = data.decode(enc["encoding"]).splitlines()
            self.t = self.i.parse(path)
        else :
            self.s = source.splitlines()
            self.t = self.i.parse(path, unsaved_files=[(path, source)])
        self.d = {}
        for child in self.t.cursor.get_children() :
            if (child.location.file.name == path
                and (child.is_definition()
                     or child.kind in (CursorKind.ENUM_CONSTANT_DECL,
                                       CursorKind.ENUM_DECL,
                                       CursorKind.FUNCTION_DECL,
                                       CursorKind.STRUCT_DECL,
                                       CursorKind.TYPEDEF_DECL,
                                       CursorKind.TYPE_ALIAS_DECL,
                                       CursorKind.UNION_DECL,
                                       CursorKind.VAR_DECL))) :
                self.d[child.spelling] = child
        self.x = self.xml()
    def __iter__ (self) :
        return iter(self.d)
    def src (self) :
        return "\n".join(self.s) + "\n"
    def add (self, src) :
        self.s.extend(src.splitlines())
    def node (self, name) :
        return self.d[name]
    def sig (self, name) :
        decl = self.d[name]
        if decl.result_type.spelling :
            return f"{decl.result_type.spelling} {decl.displayname}"
        else :
            return f"{decl.displayname}"
    def loc (self, name) :
        decl = self.d[name]
        return self.p, decl.extent.start.line, decl.extent.end.line
    def _get_slice (self, sl, sc, el, ec) :
        lines = self.s[sl-1:el]
        lines[0] = lines[0][sc-1:]
        lines[-1] = lines[-1][:ec-1]
        return "\n".join(lines)
    def __getitem__ (self, name) :
        decl = self.d[name]
        return self._get_slice(decl.extent.start.line,
                               decl.extent.start.column,
                               decl.extent.end.line,
                               decl.extent.end.column)
    def __delitem__ (self, name) :
        decl = self.d[name]
        pos = decl.extent.start.line - 1
        self.s[pos] = self.s[pos][:decl.extent.start.column - 1]
        pos = decl.extent.end.line - 1
        self.s[pos] = self.s[pos][decl.extent.end.column - 1:]
        for pos in range(decl.extent.start.line, decl.extent.end.line - 1) :
            self.s[pos] = ""
    def __call__ (self, expr, fmt="src") :
        val = self.x.xpath(expr)
        if isinstance(val, list) :
            return [self._xpath(v, fmt) for v in val]
        else :
            return self._xpath(val, fmt)
    def _xpath (self, val, fmt) :
        if isinstance(val, (int, str)) or val is None :
            return val
        elif fmt == "src" :
            return self._get_slice(int(val.attrib["startline"]),
                                   int(val.attrib["startcol"]),
                                   int(val.attrib["endline"]),
                                   int(val.attrib["endcol"]))
        elif fmt == "xml" :
            return val
        elif fmt == "node" :
            return self._x[val]
    def xml (self) :
        self._x = {}
        root = etree.Element("CTU",
                             file=self.p,
                             startline=str(self.t.cursor.extent.start.line),
                             startcol=str(self.t.cursor.extent.start.column),
                             endline=str(self.t.cursor.extent.end.line),
                             endcol=str(self.t.cursor.extent.end.column))
        for child in self.t.cursor.get_children() :
            if child.location.file.name == self.p :
                self._xml(child, root)
        return root
    def _xml (self, node, xparent) :
        xchild = etree.SubElement(xparent, node.kind.name.lower(),
                                  startline=str(node.extent.start.line),
                                  startcol=str(node.extent.start.column),
                                  endline=str(node.extent.end.line),
                                  endcol=str(node.extent.end.column))
        self._x[xchild] = node
        for name in dir(node) :
            try :
                obj = getattr(node, name)
            except :
                continue
            if name.startswith("is_") and callable(obj) :
                xchild.attrib[name] = str(obj()).lower()
            elif not obj :
                pass
            elif not name.startswith("_") :
                txt = str(obj).lower()
                if txt.startswith("<") :
                    continue
                xchild.attrib[name] = str(obj).lower()
        for child in node.get_children() :
            self._xml(child, xchild)
